/**
 * \file labirintus.cpp
 *    labirintus program.
 * $Id: labirintus.cpp 499 2015-05-05 10:53:36Z szebi $
 */

#include <iostream>
#include <stdexcept>

#include "memtrace.h"
#include "labirint.h"

using std::cout;
using std::cin;
using std::endl;



/**
 * \mainpage Labirintusban mozgó állatok modellezése
 *
 * \par Rövid specifikáció:
 *  - Tetszőleges méretű 2D labirintus definiálása.
 *  - Minden cellának 4 szomszédja van.
 *  - A logikai szomszédok nem feltétlenül egyeznek meg a geometriai szomszédokkal.
 *  - Alapobjektum tulajdonságai, műveletei:
 *    -#  jel
 *    -#  erő
 *    -#  Néz
 *    -#  Lép
 * \par Demonstrációs cél:
 *  - Visszalépéses (backtrack) algoritmus bemutatása
 *  - Objektummodell cellás játékokhoz (ötlet).
 *  - Virtuális függvények használata, lényege.
 *  - Saját generikus lista és használata.
 * \par Működés:
 *  - Az állatokat reprezentáló objektum a következő algoritmussal járja be a labirintust:
 *    - Minden cellából először jobbra, majd le, ezután balra és végül fel irányokba néz.
 *    - Ha az adott irányban levő cella üres, akkor belelép.
 *    - Ha egy cellában már van objektum, akkor az erő attribútum dönt, hogy rálép-e.
 *    - Minden cellában otthagyja a nyomát, azaz azt, hogy onnan merre indult legutoljára.
 *    - Csak olyan cellába lép be, ahol nincs saját nyoma.
 *    - Ha már nem tud hova lépni, visszalép.
 *  - Az algoritmust részletesen 12. heti előadás mutatja be
 *  - A fal ill. az üres cella is objektum, de nem tud lépni.
 * \par Megjegyzés:
 *  - A Cella objektumban az irányok trükkös kódolása helyett szebb lenne egy irány objektumot
 *    használni. Ennek egy metódusa lehetne pl. az ellenkező irányba váltás. Ez az objektum magán belül használhatna
 *    bármilyen trükköt. Így a trükk nem terjedne ki más objektumokra, másrészt a Cella belső enumj-ának használata
 *    kényelmetlen, mivel az enum nem objektum.
 *    A
 * \par Letölthető állományok:
 *  - <a href="../labirintus.cpp">labrintus.cpp</a> - főprogram fallal, egérrel, és macskával
 *  - <a href="../labirint.cpp">labirint.cpp</a> - labirint objektum megvalósítása
 *  - <a href="../cella.cpp">cella.cpp</a> - cella objektum megvalósítása
 *  - <a href="../obj.cpp">obj.cpp</a> - alapobjektum megvalósítása
 *  - <a href="../labirint.h">labirint.h</a> - labirint objektum deklarációja
 *  - <a href="../cella.h">cella.h</a> - cella objektum deklarációja
 *  - <a href="../obj.h">obj.h</a> - alapobjektum deklarációja
 *  - <a href="../lista.hpp">lista.hpp</a> - generikus lista a "nyom" kezeléséhez használjuk
 *  - <a href="../Makefile">Makefile</a> - Makefile UNIX-on történő fordításhoz
 */

/**
 * Egér objektum
 */
class Eger :public Obj {
public:
    char Jele() { return('e'); }                // 'e'-t ad vissza
    int  Ereje() { return(10); }                // egér ereje
    void operator()(Obj *p) {                   // meghívódik, ha megették
        cout << "Jaj szegeny " << Jele();
        cout << " megetvett a(z):";
        cout << p->Jele() << " jelu\n";
    }
};

/**
 * Macska objektum
 */
class Macska :public Obj {
    const int ero;
    const char jel;
public:
    Macska(char j = 'm', int e = 20) :ero(e), jel(j) {}
    char Jele() { return(jel); }                // jele
    int  Ereje() { return(ero); }               // macska ereje
    void operator()(Obj *p) {                   // meghívódik, ha megették
        cout << "Jaj szegeny " << Jele();
        cout << " megetvett a(z):";
        cout << p->Jele() << " jelu\n";
    }
};

/**
 * Fal objektum
 */
class Fal :public Obj {
public:
    char Jele() { return('#'); }                // '#'-t ad vissza
    int  Ereje() { return(1000); }              // Fal ereje
    void Lep(Cella &c) {};                      // üres függvény (nem lép)
};

/**
 * Demo program.
 *  Felépitünk egy ilyen labirintust:
 *    ..#..
 *    ...me
 *    #..e.
 *    .M...
 */
int main()
{
    Labirint l(4,5);                            // 4 * 5 os labirintus
    Eger e1, e2;                                // 2 egér; jele: e
    Macska m1, m2('M', 21);                     // 2 macska; jele: m ill. M
    Fal f1, f2;                                 // 2 fal; jele: #
    l.SetObj(2, 3, e1);                         // egér elhelyezése
    l.SetObj(1, 4, e2);                         // egér elhelyezése
    l.SetObj(1, 3, m1);                         // macska elhelyezése
    l.SetObj(3, 1, m2);                         // macska elhelyezése
    l.SetObj(0, 2, f1);                         // fal elhelyezése
    l.SetObj(2, 0, f2);                         // fal elhelyezése
    try {
      char ch = '.';
      l.Print();
      while (ch != 'c' && cin >> std::noskipws >> ch, ch != 'e') {
        l.Lep();
        l.Print();
      }
    } catch (std::exception& e) {
        cout << e.what() << endl;
    }
    return(0);
}
